﻿// // ******************************************************************
// //       /\ /|       @File         UsbDecode.cpp
// //       \ V/        @Brief
// //       | "")       @Author        lijinwen, ghz005@uni-trend.com.cn
// //       /  |        @Creation      2024-3-14
// //      /  \\        @Modified      2024-5-13
// //    *(__\_\
// // ******************************************************************

#include "UsbDecode.h"

namespace Protocol::ProtocolUsb
{
    UsbDecodeResult resultData;
    SignalRate signalRate;

    uint16_t byteCount;

    std::vector<UsbPacket> usbPackets;

    double GetBitRateByPs()
    {
        double result = 1.0;
        switch (signalRate)
        {
        case SignalRate::LowRate:
            result /= 1.5E6;
            break;
        case SignalRate::FullRate:
            result /= 12E6;
            break;
        case SignalRate::HighRate:
            result /= 480E6;
            break;
        default: break; // NOLINT(clang-diagnostic-covered-switch-default)
        }
        return result;
    }

    std::vector<std::string> EventInfoTitles()
    {
        return {"Index", "Start Time", "Sync", "PID", "Data", "Addr", "FNUM", "CRC5", "CRC16", "EOP", "Error"};
    }

    SignalRate GetSignalRate()
    {
        return signalRate;
    }

    void SetSignalRate(SignalRate value)
    {
        signalRate = value;
    }

    uint16_t ByteCount()
    {
        return byteCount;
    }

    void SetByteCount(uint16_t value)
    {
        byteCount = value;
    }

    uint16_t MaxByteCount()
    {
        return 1023;
    }

    uint16_t MinByteCount()
    {
        return 0;
    }

    bool UsbDecode::ParseData(UsbDecodeParams options, std::vector<TwoLevelEdgePulse> edgePluses1,
                              std::vector<TwoLevelEdgePulse> edgePluses2) const
    {
        if (edgePluses1.empty()) return false;
        if (options.SignType == SignType::Diff && edgePluses2.empty()) return false;

        //todo 差分逻辑处理

        // 检查是否取消
        if (canceled_)
        {
            std::cout << "AsyncFunction canceled." << '\n';
            return false;
        }
        int32_t leftDataSize = static_cast<int32_t>(edgePluses1.size());
        if (FindPackets(options, edgePluses1.data(), leftDataSize, usbPackets))
        {
            if (usbPackets.empty())
            {
                return false;
            }
            resultData.ClearData();

            //todo
            for (auto&& element : usbPackets)
            {
                resultData.Add(new UsbDecodeResultUnit(element));
            }
        }

        return true;
    }

    bool UsbDecode::CheckDifferenceWithinThreshold(const double* array, int32_t size, double threshold)
    {
        //threshold = 0.15;
        for (int i = 1; i < size; i++)
        {
            const double difference = std::abs(array[i] - array[i - 1]);
            const double percentageDifference = difference / array[i - 1];
            if (difference < 50)
            {
                threshold = 0.28;
            }
            if (percentageDifference > threshold)
            {
                return false; // 如果差异大于阈值，则返回 false
            }
        }

        return true; // 所有元素之间的差异都小于阈值
    }

    void UsbDecode::ClearEdgeArray(Edge* edges, int32_t size)
    {
        for (int i = 0; i < size; ++i)
        {
            edges[i] = Edge::None;
        }
    }

    bool UsbDecode::CheckSyncEdges(Edge startEdge, const Edge* edges)
    {
        return edges[0] == edges[1] && edges[0] == startEdge && edges[0] == edges[6];
    }

    bool UsbDecode::CheckSyncSpans(const TwoLevelEdgePulse* node, int32_t& leftOverSize
                                   , const int32_t count, int32_t& avgSpans)
    {
        if (count < 1)
        {
            return false;
        }
        const auto startIndex = new int32_t[count + 1];
        avgSpans = 0;
        if ((node - 1) != nullptr)
        {
            node--;
        }

        for (int i = count; i >= 0; i--)
        {
            if (node == nullptr)
            {
                return false;
            }
            startIndex[i] = node->StartIndex;
            node--;
        }
        return CheckSyncSpansByArray(startIndex, count, avgSpans);
    }

    bool UsbDecode::CheckSyncSpansByArray(const int32_t* startIndex, const int32_t count,
                                          int32_t& avgSpans)
    {
        if (count < 1)
        {
            return false;
        }
        const auto tmpSpans = new double[count - 1];
        for (int i = 1; i < count; i++)
        {
            tmpSpans[i - 1] = static_cast<double>(startIndex[i] - startIndex[i - 1]);
        }
        double totalSpans = 0;
        for (int i = 0; i < count - 1; i++)
        {
            totalSpans += std::abs(tmpSpans[0]);
        }
        avgSpans = static_cast<int>(totalSpans / (count - 1));
        return CheckDifferenceWithinThreshold(tmpSpans, count - 1);
    }

    bool UsbDecode::FindAllSyncs(TwoLevelEdgePulse* node, int32_t& leftOverSize,
                                 std::vector<SYNC>& allSYNC,
                                 bool polarity) const
    {
        allSYNC.clear();
        int firstAvgSpan = 0;

        if (node == nullptr || node - 1 == nullptr)
        {
            return false;
        }

        constexpr int syncBitLen = 8;
        const Edge nextEdge = polarity ? Edge::Rise : Edge::Falling;
        const Edge startEdge = nextEdge;

        TwoLevelEdgePulse* startInfoNode = nullptr;
        Edge tmpEdges[syncBitLen];
        ClearEdgeArray(tmpEdges, syncBitLen);

        int tmpEdgeCount = 0;
        TwoLevelEdgePulse* tmpNode = node;

        while (tmpNode != nullptr)
        {
            // 检查是否取消
            if (canceled_)
            {
                std::cout << "AsyncFunction canceled." << '\n';
                return false;
            }
            if (tmpEdgeCount <= 7)
            {
                if (tmpEdgeCount == 0)
                {
                    startInfoNode = tmpNode;
                }

                tmpEdges[tmpEdgeCount] = tmpNode->Edge;
                tmpEdgeCount++;
            }
            else
            {
                if (CheckSyncEdges(startEdge, tmpEdges))
                {
                    int avgSpans;
                    if (CheckSyncSpans(tmpNode, leftOverSize, syncBitLen - 1, avgSpans))
                    {
                        if (allSYNC.empty())
                        {
                            firstAvgSpan = avgSpans;
                            allSYNC.emplace_back(startInfoNode->StartIndex, (tmpNode - 1)->StartIndex,
                                                 avgSpans);
                        }
                        else if (startInfoNode != nullptr && (startInfoNode - 1) != nullptr && (startInfoNode -
                                1)->
                            GetLength() > firstAvgSpan * 8)
                        {
                            allSYNC.emplace_back(startInfoNode->StartIndex, (tmpNode - 1)->StartIndex,
                                                 avgSpans);
                        }

                        ClearEdgeArray(tmpEdges, syncBitLen);
                        tmpEdgeCount = 0;
                        tmpNode++;
                        leftOverSize--;
                        continue;
                    }
                }

                for (int i = 0; i < syncBitLen - 1; i++)
                {
                    tmpEdges[i] = tmpEdges[i + 1];
                }

                tmpEdges[syncBitLen - 1] = tmpNode->Edge;
                startInfoNode++;
            }

            tmpNode++;
            leftOverSize--;
        }

        return !allSYNC.empty();
    }

    bool UsbDecode::FindPackets(UsbDecodeParams options, TwoLevelEdgePulse* node, int32_t& leftOverSize,
                                std::vector<UsbPacket>& allPacket,
                                bool polarity) const
    {
        allPacket.clear();
        std::vector<SYNC> allSYNC;
        // 检查是否取消
        if (canceled_)
        {
            std::cout << "AsyncFunction canceled." << '\n';
            return false;
        }


        TwoLevelEdgePulse* tmpNode = node;
        if (!FindAllSyncs(tmpNode, leftOverSize, allSYNC, polarity))
        {
            std::cout << "SYNC not found." << '\n';
            return false;
        }

        tmpNode = node;
        for (const auto& sync : allSYNC)
        {
            while (tmpNode != nullptr && tmpNode->StartIndex < sync.StartIndex)
            {
                // 检查是否取消
                if (canceled_)
                {
                    std::cout << "AsyncFunction canceled." << '\n';
                    return false;
                }
                tmpNode--;
            }

            if (tmpNode == nullptr || tmpNode - 1 == nullptr)
            {
                break;
            }

            if (sync.StartIndex != tmpNode->StartIndex)
            {
                continue;
            }

            while (sync.EndIndex > tmpNode->StartIndex && tmpNode != nullptr)
            {
                tmpNode++;
                // 检查是否取消
                if (canceled_)
                {
                    std::cout << "AsyncFunction canceled." << '\n';
                    return false;
                }
            }

            if (tmpNode != nullptr)
            {
                UsbPacket packet(tmpNode, leftOverSize, sync, polarity);
                if (packet.IsValid())
                {
                    allPacket.push_back(packet);
                }
            }
        }

        return true;
    };

    bool UsbDecode::ParseData(const UsbDecodeParams& options, const TwoLevelEdgePulse& edgePluses1,
                              const TwoLevelEdgePulse& edgePluses2, UsbDecodeResult& decodeResult)
    {
        //todo
        return false;
    }

    UsbDecode::~UsbDecode()
    {
        //todo
    }

    void UsbDecode::Decode(const DecodeParams& decodeParams)
    {
        // if (dynamic_cast<const UsbDecodeParams*>(&decodeParams) == nullptr)
        // {
        //     status_ = ProtocolStatus::DecodeFailure;
        //     return;
        // }
        // const UsbDecodeParams options = dynamic_cast<const UsbDecodeParams*>(&decodeParams)[0];

        //todo
        // const TwoLevelEdgePulse* edgePulse1 = options.EdgePulses1;
        // const TwoLevelEdgePulse* edgePulse2 = options.EdgePulses2;
        // const bool result = ParseData(options, edgePulse1, edgePulse2);
        //
        // if (!result)
        // {
        //     status_ = ProtocolStatus::DecodeFailure;
        // }
    }

    void UsbDecode::QuantizeDecode(const QuantizeParams& decodeParams)
    {
        //todo 整形
    }

    DecodeResult* UsbDecode::GetResult()
    {
        //return &resultData;
        //todo
        return nullptr;
    }
}
