﻿// // ******************************************************************
// //       /\ /|       @File         UsbPacket.cpp
// //       \ V/        @Brief
// //       | "")       @Author        lijinwen, ghz005@uni-trend.com.cn
// //       /  |        @Creation      2024-3-14
// //      /  \\        @Modified      2024-5-13
// //    *(__\_\
// // ******************************************************************

#include "UsbPacket.h"

namespace Protocol::ProtocolUsb
{
    UsbPacket::UsbPacket(): pid(0), lastNRBZDataBit(false), polarity(false), endPoint(0), crc5(0), isValid(false),
                            isCRCChecked(false)
    {
    }

    UsbPacket::UsbPacket(TwoLevelEdgePulse* pidNode, int32_t& leftOverSize, SYNC sync, bool polarity)
    {
        this->polarity = polarity;
        this->packetSYNC = sync;
        //this->pidType = nullptr;

        if (!GetPID(pidNode, leftOverSize))
        {
            return;
        }

        isValid = true;

        if (this->type == UsbPacketType::NoDefine)
        {
            return;
        }

        DecodeFields(pidNode, leftOverSize, sync);
    }

    bool UsbPacket::LastNRBZDataBit() const
    {
        return lastNRBZDataBit;
    }

    bool UsbPacket::CheckNibblesInverse(uint8_t value)
    {
        const uint8_t highNibble = static_cast<uint8_t>(value >> 4);
        const uint8_t lowNibble = static_cast<uint8_t>(value & 0x0F);
        return highNibble == (~lowNibble & 0x0F);
    }

    bool UsbPacket::GetPID(TwoLevelEdgePulse* pidNode, int32_t& leftOverSize)
    {
        if (pidNode == nullptr || leftOverSize <= 0 || packetSYNC.SingleBitTimingLength <= 0)
        {
            return false;
        }
        lastNRBZDataBit = true;
        uint8_t data;
        if (!DecodeNextByte(pidNode, leftOverSize, data))
        {
            return false;
        }
        // if (pidNode != nullptr) {
        //     pidEndIndex = pidNode.Parent().StartIndex();
        // }

        pid = data;

        if (CommonHelper::EnumIsDefined<TokenPackageType>(
            RealPID()))
        {
            type = UsbPacketType::Token;
        }
        else if (CommonHelper::EnumIsDefined<DataPackageType>(
            RealPID()))
        {
            type = UsbPacketType::Data;
        }
        else if (CommonHelper::EnumIsDefined<HandshakePackageType>(
            RealPID()))
        {
            type = UsbPacketType::HandShake;
        }
        else if (CommonHelper::EnumIsDefined<SpecialPacketType>(
            RealPID()))
        {
            type = UsbPacketType::Special;
        }
        else
        {
            return false;
        }
        return true;
    }

    bool UsbPacket::DecodeNextByte(TwoLevelEdgePulse* node, int32_t& leftOverSize, uint8_t& outData)
    {
        outData = 0;
        const bool highPolarityEdge = polarity;
        const double singleBitTimingLengthThreshold = packetSYNC.SingleBitTimingLength * 0.95;
        int gotbitCount = 0;
        std::vector<bool> nrzi_data(8, false);

        while (gotbitCount < 8)
        {
            if (node == nullptr || leftOverSize <= 0 || node->Edge == Edge::None)
            {
                return false;
            }

            if (node->EndIndex - node->StartIndex >= singleBitTimingLengthThreshold)
            {
                const int bitsCount = static_cast<int>((node->EndIndex - node->StartIndex) /
                    singleBitTimingLengthThreshold);

                const int lastGotbitCount = gotbitCount;
                for (int i = 0; i < bitsCount; i++)
                {
                    if (lastGotbitCount + i >= 8)
                    {
                        continue;
                    }
                    nrzi_data[lastGotbitCount + i] = highPolarityEdge == (node->Edge ==
                        Edge::Rise);

                    gotbitCount++;
                }
            }
            leftOverSize--;
            node++;
        }

        for (int i = 0; i < 8; i++)
        {
            if (lastNRBZDataBit != nrzi_data[i])
            {
                lastNRBZDataBit = nrzi_data[i];
            }
            else
            {
                outData |= (1 << i);
            }
        }

        return true;
    }

    void UsbPacket::DecodeFields(TwoLevelEdgePulse* node, int32_t& leftOverSize, const SYNC& sync)
    {
        if (leftOverSize <= 0)
        {
            return;
        }
        switch (type)
        {
        case UsbPacketType::Token:
            DecodeFieldsByToken(node, leftOverSize, sync);
            break;
        case UsbPacketType::Data:
            DecodeFieldsByData(node, leftOverSize, sync);
            break;
        case UsbPacketType::HandShake:
            DecodeFieldsByHandShake(node, leftOverSize, sync);
            break;
        case UsbPacketType::Special:
            DecodeFieldsBySpecial(node, leftOverSize, sync);
            break;
        case UsbPacketType::NoDefine:
            break;
        }
    }

    void UsbPacket::DecodeFieldsByToken(TwoLevelEdgePulse* node, int32_t& leftOverSize, const SYNC& sync)
    {
        if (type == UsbPacketType::NoDefine)
        {
            return;
        }

        //todo
        switch (const auto packageType = static_cast<
            TokenPackageType>(RealPID()))
        {
        case TokenPackageType::Out:
            break;
        case TokenPackageType::SOF:
            break;
        case TokenPackageType::In:
            break;
        case TokenPackageType::Setup:
            break;
        default:
            break;
        }
    }

    void UsbPacket::DecodeFieldsByData(TwoLevelEdgePulse* node, int32_t& leftOverSize, const SYNC sync)
    {
        if (type == UsbPacketType::NoDefine)
        {
            return;
        }
        switch (const auto packageType = static_cast<DataPackageType>
            (RealPID()))
        {
        case DataPackageType::Data0:
        case DataPackageType::Data1:
            datas.clear();
            while (node != nullptr && node->StartIndex < sync.PacketEndIndex())
            {
                uint8_t data;
                DecodeNextByte(node, leftOverSize, data);
                datas.push_back(data);
            }
            break;
        case DataPackageType::Data2:
            break;
        case DataPackageType::MData:
            break;
        default:
            break;
        }
    }

    void UsbPacket::DecodeFieldsByHandShake(TwoLevelEdgePulse* node, int32_t& leftOverSize, SYNC sync) const
    {
        //todo
        if (type != UsbPacketType::HandShake)
        {
        }
    }

    void UsbPacket::DecodeFieldsBySpecial(TwoLevelEdgePulse* node, int32_t& leftOverSize, SYNC sync) const
    {
        //todo
        if (type != UsbPacketType::Special)
        {
        }
    }
}
